/* $Id$ */
/**************************************************************************
 *
 *  FILE            network.c
 *  MODULE OF       vidclient
 *
 *  DESCRIPTION     
 *
 *  WRITTEN BY      Sverre H. Huseby <sverrehu@ifi.uio.no>
 *
 **************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <shhmsg.h>
#include <shhnet.h>

#include "protocol.h"
#include "blockimg.h"
#include "shhsched.h"
#include "window.h"
#include "vidclient.h"
#include "network.h"

/**************************************************************************
 *                                                                        *
 *                       P R I V A T E    D A T A                         *
 *                                                                        *
 **************************************************************************/

static unsigned char netInPacket[MAX_PACKET];
static int netInIndex;

static char netRemoteSystem[81] = "";
static char netRemoteHost[81];
static char netRemoteTcpPort[81];
static int  netRemoteUdp = -1;
static int  netRemoteTcp = -1;

static int  netLocalUdp;

static SchedId netTcpReadyId;
static SchedId netUdpReadyId;

static BlockImage blkImgIn;



/**************************************************************************
 *                                                                        *
 *                   P R I V A T E    F U N C T I O N S                   *
 *                                                                        *
 **************************************************************************/

static void
netHangup(void)
{
    unsigned char buff[128];

    if (netRemoteTcp < 0)
	return;
    msgMessage("sending hangup to %s:%s\n", netRemoteHost, netRemoteTcpPort);
    buff[0] = PACK_HANGUP;
    tcpWriteAll(netRemoteTcp, buff, 1);
    vidStopRepaint();
}

static void
netUdpInput(void *clientData, int fd)
{
    int                    n, type, len, x, y, nBytes;
#if GRAY_BITS == 4
    register unsigned char *s, *d;
    register int           q;
#endif

    udpRead(fd, netInPacket, sizeof(netInPacket), &n);
    if (n < 1) {
	msgError("empty packet received\n");
	return;
    }
#if GRAY_BITS == 4
    nBytes = (BLOCK_WIDTH * BLOCK_HEIGHT) / 2;
#else
    nBytes = BLOCK_WIDTH * BLOCK_HEIGHT;
#endif
    netInIndex = 0;
    while (netInIndex < n) {
	type = netInPacket[netInIndex++];
	len = n - netInIndex;
	switch (type) {
	  case PACK_BLOCK:
	    if (len < 2) {
		msgError("partial block received\n");
		goto finish;
	    }
	    x = netInPacket[netInIndex++];
	    y = netInPacket[netInIndex++];
#if GRAY_BITS == 4
	    s = netInPacket + netInIndex;
	    d = blkImgIn.block[x][y].data;
	    for (q = nBytes; q; q--) {
		*d++ = *s >> 4;
		*d++ = *s++ & 0x0F;
	    }
#else
	    memcpy(blkImgIn.block[x][y].data, netInPacket + netInIndex,
		   nBytes);
#endif
	    netInIndex += nBytes;
	    winNewBlock(x, y, blkImgIn.block[x][y].data);
	    break;
	  default:
	    msgError("unknown packet received\n");
	}
    }
  finish:
}

static void
netTcpInput(void *clientData, int fd)
{
    unsigned char buff[128];
    int           n, type;
    SNAddr        addr;
    SNPort        port;

    if (netRemoteTcp < 0)
	netRemoteTcp = fd;
    tcpRead(fd, buff, sizeof(buff), &n);
    if (n == 0) {
	msgError("got empty tcp packet, assuming broken connection\n");
	schedEndLoop();
	return;
    }
    type = buff[0];
    switch (type) {
      case PACK_BUSY:
	msgError("server busy, shutting down.\n");
	schedEndLoop();
	break;
      case PACK_SEND_PORT:
	msgMessage("server wants our UDP port. creating one.\n");
	udpClientSock(&netLocalUdp);
	netUdpReadyId
	    = schedAddDesc(SCHED_READ_READY, netLocalUdp, netUdpInput, NULL);
	inGetSockAddrPort(netLocalUdp, &addr, &port);
	buff[0] = PACK_PORT;
	buff[1] = (port >> 8) & 0xFF;
	buff[2] = port & 0xFF;
	tcpWriteAll(fd, buff, 3);
	break;
      case PACK_WELCOME:
	msgMessage("we're in. setting up repainting of window\n");
	vidStartRepaint();
	break;
      case PACK_GO_AWAY:
	msgError("server doesn't want us, shutting down.\n");
	schedEndLoop();
	break;
      default:
	msgError("unknown packet received (%u)\n", (unsigned) type);
    }
}

static void
netCallRemote(void)
{
    msgMessage("calling %s:%s\n", netRemoteHost, netRemoteTcpPort);
    tcpClientOpen(netRemoteHost, netRemoteTcpPort, &netRemoteTcp);
    netTcpReadyId
	= schedAddDesc(SCHED_READ_READY,
		       netRemoteTcp, netTcpInput, NULL);
}

static void
netGetHostAndPort(const char *system, char *host, char *port)
{
    const char *p;

    if ((p = strchr(system, ':')) == NULL) {
	strcpy(host, system);
	sprintf(port, "%d", DEFAULT_PORT);
    } else {
	if (p == system)
	    strcpy(host, "localhost");
	else
	    strncpy(host, system, p - system);
	strcpy(port, p + 1);
	if (!strlen(port))
	    sprintf(port, "%d", DEFAULT_PORT);
    }
}



/**************************************************************************
 *                                                                        *
 *                    P U B L I C    F U N C T I O N S                    *
 *                                                                        *
 **************************************************************************/

void
netSetRemoteSystem(const char *system)
{
    strcpy(netRemoteSystem, system);
}

void
netInit(void)
{
    netInIndex = 0;

    netGetHostAndPort(netRemoteSystem, netRemoteHost, netRemoteTcpPort);
    netCallRemote();
}

void
netFinish(void)
{
    netHangup();

    if (netLocalUdp >= 0) {
	udpClose(netLocalUdp);
	netLocalUdp = -1;
    }
    if (netRemoteTcp >= 0) {
	tcpClose(netRemoteTcp);
	netRemoteTcp = -1;
	schedRemoveDesc(netTcpReadyId);
    }
    if (netRemoteUdp >= 0) {
	udpClose(netRemoteUdp);
	netRemoteUdp = -1;
	schedRemoveDesc(netUdpReadyId);
    }
}
