/* $Id$ */
/**************************************************************************
 *
 *  FILE            window.c
 *  MODULE OF       vidclient
 *
 *  DESCRIPTION     
 *
 *  WRITTEN BY      Sverre H. Huseby <sverrehu@ifi.uio.no>
 *
 **************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <values.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>  /* XVisualInfo and friends */
#ifdef USE_SHM
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
/* My XShm.h does not declare the following: */
int XShmQueryExtension(Display *);
int XShmGetEventBase(Display *);
#endif

#include <shhmsg.h>

#include "protocol.h"
#include "shhsched.h"
#include "window.h"

#undef DRAW_BOX

/**************************************************************************
 *                                                                        *
 *                       P R I V A T E    D A T A                         *
 *                                                                        *
 **************************************************************************/

static char    *winDisplayName = NULL;
static int      winScreenNum;
static Display *winDisplay;
static Window   winParent = 0;
static int      winDepth;
static int      winX = 0, winY = 0;
static int      winWidth = IMAGE_WIDTH, winHeight = IMAGE_HEIGHT;
static Window   winOut;
static Colormap winColormap;
static GC       winGC;
static unsigned long winGray[NUM_GRAY];
static XImage  *winImage;
static unsigned char *winBuff;
#ifdef USE_SHM
static int      winShm = 1;
static XShmSegmentInfo winShmInfo;
static int      winShmErrorFlag;
#endif



/**************************************************************************
 *                                                                        *
 *                   P R I V A T E    F U N C T I O N S                   *
 *                                                                        *
 **************************************************************************/

static int errBadValue;
static int (*oldBadValueHandler)(Display *, XErrorEvent *);

static int
xBadValueHandler(Display *dpy, XErrorEvent *err)
{
    if (err->error_code == BadValue)
	errBadValue = 1;
    else if (oldBadValueHandler)
	oldBadValueHandler(dpy, err);
    return 0;
}

#define KEEP
static Status
xAllocClosestColor(Display *dpy, Colormap cm, XColor *xc)
{
#define MAX_COL 512
#define KEEP
    int    q;
    Status ret;
#ifdef KEEP
    static int      nCol = 0;
    static XColor   *color = NULL;
    static double   *dist = NULL;
    static Colormap lastMap = 0;
#else
    int    nCol;
    XColor *color;
    double *dist;
#endif
    double rdiff, gdiff, bdiff, best;
    int    bestIdx;

    /* first try to do a normal allocation */
    if ((ret = XAllocColor(dpy, cm, xc)) != 0)
	return ret; /* ok, that worked! */

    /* that didn't work. */
    /* try allocating the closest among the first MAX_COL available colors */
#ifdef KEEP
    if (!color && (color = malloc(MAX_COL * sizeof(XColor))) == NULL)
	return 0;
    if (!dist && (dist = malloc(nCol * sizeof(double))) == NULL)
	return 0;
#else
    if ((color = malloc(MAX_COL * sizeof(XColor))) == NULL)
	return 0;
    if ((dist = malloc(nCol * sizeof(double))) == NULL) {
	free(color);
	return 0;
    }
#endif

    /* first fetch the colors */
#ifdef KEEP
    if (cm != lastMap) {
	lastMap = cm;
#endif
	errBadValue = 0;
	oldBadValueHandler = XSetErrorHandler(xBadValueHandler);
	for (nCol = 0; nCol < MAX_COL; nCol++) {
	    color[nCol].pixel = nCol;
	    XQueryColor(dpy, cm, &color[nCol]);
	    if (errBadValue)
		break;
	}
	XSetErrorHandler(oldBadValueHandler);
#ifdef KEEP
    }
#endif

    /* build table of `distances' */
    for (q = 0; q < nCol; q++) {
	rdiff = xc->red - color[q].red;
	rdiff *= rdiff;
	gdiff = xc->green - color[q].green;
	gdiff *= gdiff;
	bdiff = xc->blue - color[q].blue;
	bdiff *= bdiff;
	dist[q] = rdiff + gdiff + bdiff;
    }
    /* then find the closest one, and allocate it. */
    /* it may be that the alloc fails. in that case, find the next closest */
    do {
	best = MAXDOUBLE;
	bestIdx = -1;
	for (q = 0; q < nCol; q++)
	    if (dist[q] < best) {
		best = dist[q];
		bestIdx = q;
	    }
	if (best < MAXDOUBLE) {
	    /* try to allocate */
	    if ((ret = XAllocColor(dpy, cm, &color[bestIdx])) != 0) {
		memcpy(xc, &color[bestIdx], sizeof(XColor));
		goto finish;
	    } else {
		dist[bestIdx] = MAXDOUBLE; /* non-allocatable */
	    }
	}
    } while (best < MAXDOUBLE);

  finish:
#ifndef KEEP
    free(dist);
    free(color);
#endif

    return ret;
}

static void
winOutCreate(void)
{
    winOut = XCreateSimpleWindow(winDisplay, winParent,
				 winX, winY,
				 winWidth, winHeight,
				 0,
				 WhitePixel(winDisplay, winScreenNum),
				 BlackPixel(winDisplay, winScreenNum));
    XMapWindow(winDisplay, winOut);
    XSelectInput(winDisplay, winOut, ExposureMask | ButtonPressMask);
}

static void
winOutSetup(void)
{
    int q;
    XColor xc;

    for (q = 0; q < NUM_GRAY; q++) {
	xc.red = xc.green = xc.blue = q * 65535L / (NUM_GRAY - 1);
	if (xAllocClosestColor(winDisplay, winColormap, &xc) == 0)
	    msgError("unable to allocate color\n");
	winGray[q] = xc.pixel;
    }
}

static void
winOutShutdown(void)
{
    XFreeColors(winDisplay, winColormap, winGray, NUM_GRAY, 0);
}

#ifdef USE_SHM
static int
winShmErrorHandler(Display *dpy, XErrorEvent *xe)
{
    winShmErrorFlag = 1;
    return 0;
}
#endif

static void
winImageCreate(void)
{
    int   nPixel;
    short endianTest = 1;

    nPixel = winWidth * winHeight;
#ifdef USE_SHM
    if (winShm) {
	winImage = XShmCreateImage(winDisplay, None, winDepth, ZPixmap,
				   NULL, &winShmInfo, winWidth, winHeight);
	if (winImage == NULL) {
	    msgError("unable to create shared memory image\n");
	    goto no_shm;
	}
	winShmInfo.shmid
            = shmget(IPC_PRIVATE, winImage->bytes_per_line * winHeight,
                     IPC_CREAT | 0777);
        if (winShmInfo.shmid < 0) {
	    msgPerror("shmget");
            XDestroyImage(winImage);
            winShm = 0;
            goto no_shm;
        }
        winBuff = winImage->data = winShmInfo.shmaddr
	    = (char *) shmat(winShmInfo.shmid, 0, 0);
        if (winShmInfo.shmaddr == (char *) -1) {
	    msgPerror("shmat\n");
            XDestroyImage(winImage);
            winShm = 0;
            goto no_shm;
        }
        winShmInfo.readOnly = 0;
	winShmErrorFlag = 0;
	XSetErrorHandler(winShmErrorHandler);
        XShmAttach(winDisplay, &winShmInfo);
        XSync(winDisplay, 0);
	XSetErrorHandler(NULL);
	if (winShmErrorFlag) {
	    msgError("can't use shared memory on remote display\n");
	    XFlush(winDisplay);
	    XDestroyImage(winImage);
	    shmdt(winShmInfo.shmaddr);
	    shmctl(winShmInfo.shmid, IPC_RMID, 0);
            winShm = 0;
            goto no_shm;
 	}
	shmctl(winShmInfo.shmid, IPC_RMID, 0);
    }
  no_shm:
    if (!winShm) {
#endif
	if ((winBuff = malloc(nPixel)) == NULL)
	    msgFatal("out of memory\n");
	winImage = XCreateImage(winDisplay, None, winDepth, ZPixmap,
				0, winBuff, winWidth, winHeight, 8, 0);
#ifdef USE_SHM
    }
#endif
    if (*(char *)&endianTest) {
	winImage->byte_order = LSBFirst;
	winImage->bitmap_bit_order = LSBFirst;
    } else {
	winImage->byte_order = MSBFirst;
	winImage->bitmap_bit_order = MSBFirst;
    }
}

static void
winImageShutdown(void)
{
#ifdef USE_SHM
    if (winShm)
	XShmDetach(winDisplay, &winShmInfo);
#endif
    XDestroyImage(winImage);
#ifdef USE_SHM
    if (winShm)
	shmdt(winShmInfo.shmaddr);
#endif
}

static void
winEventReady(void *clientData, int fd)
{
    XEvent xe;

    XNextEvent(winDisplay, &xe);
    switch (xe.type) {
      case Expose:
	if (xe.xexpose.count)
	    break;
	winRepaint();
	break;
      case ButtonPress:
	schedEndLoop();
	break;
      default:
    }
}



/**************************************************************************
 *                                                                        *
 *                    P U B L I C    F U N C T I O N S                    *
 *                                                                        *
 **************************************************************************/

void
winSetParentWindow(const char *s)
{
    winParent = (Window) strtol(s, NULL, 0);
}

void
winInit(void)
{
    int px, py;
    unsigned pw, ph, pbord, pdepth;
    Window proot;
    XWindowAttributes wa;

    if ((winDisplay = XOpenDisplay(winDisplayName)) == NULL)
	msgFatal("cannot connect to X server for display `%s'\n",
		 XDisplayName(winDisplayName));
    winScreenNum = DefaultScreen(winDisplay);
    if (!winParent) {
	winParent = DefaultRootWindow(winDisplay);
	winColormap = DefaultColormap(winDisplay, winScreenNum);
    } else {
	XGetGeometry(winDisplay, winParent,
		     &proot, &px, &py, &pw, &ph, &pbord, &pdepth);
printf("parent: %u, %u\n", pw, ph);
	if ((winX = ((int) pw - winWidth) / 2) < 0)
	    winX = 0;
	if ((winY = ((int) ph - winHeight) / 2) < 0)
	    winY = 0;
	if (XGetWindowAttributes(winDisplay, winParent, &wa) == 0
	    || wa.colormap == None)
	    winColormap = DefaultColormap(winDisplay, winScreenNum);
	else
	    winColormap = wa.colormap;
    }
    winGC = DefaultGC(winDisplay, winScreenNum);
    winDepth = DefaultDepth(winDisplay, winScreenNum);
#ifdef USE_SHM
    if (winShm && XShmQueryExtension(winDisplay) == 0) {
	msgError("no shared memory extensions present\n");
	winShm = 0;
    }
#endif

    winOutCreate();
    winOutSetup();
    winImageCreate();

    schedAddDesc(SCHED_READ_READY, ConnectionNumber(winDisplay),
		 winEventReady, NULL);
}

void
winFinish(void)
{
    winImageShutdown();
    winOutShutdown();
    XCloseDisplay(winDisplay);
}

void
winNewBlock(int x, int y, unsigned char *data)
{
    register int q;
    register unsigned char *where;
    unsigned char *start;
    int      w, rx, ry;

    start = winBuff
	      + (ry = y * BLOCK_HEIGHT) * winWidth
	      + (rx = x * BLOCK_WIDTH);
    for (w = 0; w < BLOCK_HEIGHT; w++) {
	where = start + w * winWidth;
	for (q = BLOCK_WIDTH; q; q--)
	    *where++ = winGray[*data++];
    }
#ifdef DRAW_BOX
    XDrawRectangle(winDisplay, winOut, winGC,
		   rx, ry, BLOCK_WIDTH + 1, BLOCK_HEIGHT + 1);
    XFlush(winDisplay);
#endif
}

void
winRepaint(void)
{
#ifdef USE_SHM
    if (winShm)
	XShmPutImage(winDisplay, winOut, winGC,
		     winImage, 0, 0, 0, 0, winWidth, winHeight, 0);
    else
#endif
	XPutImage(winDisplay, winOut, winGC,
		  winImage, 0, 0, 0, 0, winWidth, winHeight);
    XFlush(winDisplay);
}
